local Chest = require("storage.chest")
local ItemGroup = require("storage.itemgroup")
local Storage = require("storage")
local Text = require("gfx.text")
local List = require("gfx.list")
local Padding = require("gfx.padding")


local CACHE_FILE = "/.storage.cache"
LOCK_FILE = "/.storage.lock"

local function lock()
  if fs.exists(LOCK_FILE) then
    return false
  end

  -- Create lock file
  return fs.open(LOCK_FILE, "w+").close()
end

local function unlock()
  if not fs.exists(LOCK_FILE) then
    return false
  end

  -- Delete lock file
  local result, reason = pcall(fs.delete, LOCK_FILE)
  if not result then
    print("ERROR: Lock file could not be deleted: " .. reason)
  end

  return result
end

local function isLocked()
  return fs.exists(LOCK_FILE)
end

local function saveState(fname, ctrl)
  local ser = ctrl:toSerializable()
  local file = fs.open(fname, "w+")
  file.write(textutils.serialize(ser))
  file.close()
end

local function loadState(fname, node)
  if not isLocked() then
    print("Not locked! Loading cache...")
    local file = fs.open(fname, "r")
    if file ~= nil then
        local ser = textutils.unserialize(file.readAll())
        file.close()
        return Storage:fromSerializable(ser)
    end
  end
  
  print("Controller must scan chests...")
  local nodeName = peripheral.getName(node)
  local storageChests = {peripheral.find("inventory")}
  for i,v in ipairs(storageChests) do
    if peripheral.getName(v) == nodeName then
      table.remove(storageChests, i)
      break
    end
  end
  local controller = Storage:fromPeripherals(Storage.assumeHomogeneous(storageChests))
  saveState(fname, controller)
  unlock()
  return controller
end


local function itemList(groups, wBudget, onClick)
  local bgColors = {
    colors.gray,
    colors.black
  }
  local entries = {}

  for i=1,#groups do
    local group = groups[i]
    local text = group:getDisplayName()
    local count = tostring(group:getItemCount())

    -- Fit text inside of width budget
    local countLen = #count
    if countLen + 2 > wBudget then
      error("Width budget is too small")
    end

    local textBudget = wBudget - 2 - countLen
    if #text > textBudget then
      -- Truncate to available budget
      text = text:sub(1,textBudget)
    end

    local textLabel = Text:new{ text = text, bgColor = bgColors[i % 2] }
    local countLabel = Text:new{ text = count, bgColor = bgColors[i % 2] }
    local paddedText = Padding:new{
      left = 0,
      right = wBudget - #count - #text,
      top = 0,
      bottom = 0,
      bgColor = bgColors[i % 2],
      element = textLabel
    }

    local list = List:new{
      children = {
        paddedText,
        countLabel
      },
      vertical = false,
      onClick = onClick
    }
    table.insert(entries, list)
  end

  return List:new{
    children = entries,
    vertical = true
  }
end

local function renderDefault(state, rootElement)
  local event = {os.pullEvent()}
  local handled = rootElement:handleEvent(event)
  if not handled and event[1] == "monitor_resize" then
    local width, height = state.monitor.getSize()
    if state.width ~= width or state.height ~= height then
      state.width = width
      state.height = height

      -- Trigger re-render
      rootElement:setDirty(true)
    end
  end

  if not handled then
    os.queueEvent(table.unpack(event))
  end
end

local PAGES = {
  MAIN = function(state)
    local found = ItemGroup.collectStacks(state.controller:find(function(stack)
      return not stack:isEmpty()
    end))

    table.sort(found, function(a, b) return a:getItemCount() > b:getItemCount() end)

    local subset = {}
    for i=1,math.min(13, #found) do
      table.insert(subset, found[i])
    end

    local listResult = itemList(subset, state.width, function(group, x, y, source)
      print("Clicked: "..group:getDisplayName())
    end)
    listResult:setParent(state.monitor)

    return function()
      renderDefault(state, listResult)
    end
  end,

  GROUP_DETAIL = function(state)

  end,

  REQUEST = function(state)

  end
}


local accessNode = Chest:fromPeripheral("minecraft:trapped_chest", true)
---@diagnostic disable-next-line: need-check-nil
local controller = loadState(CACHE_FILE, accessNode:getInventory())
local monitor = peripheral.find("monitor")
local width, height = monitor.getSize()

local CONTROLLER_STATE = {
  controller = controller,
  width = width,
  height = height,
  monitor = monitor,
  currentPage = "MAIN",
  exit = false
}

os.queueEvent("dummy_event")
while not CONTROLLER_STATE.exit do
  PAGES[CONTROLLER_STATE.currentPage](CONTROLLER_STATE)

  if CONTROLLER_STATE.nextPage ~= nil then
    CONTROLLER_STATE.currentPage = CONTROLLER_STATE.nextPage
    CONTROLLER_STATE.nextPage = nil
  end
  
  sleep(0)
end